//---------------------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <conio.h>

//Includes fr PKCS#11
#include "cryptoki.h"

#pragma hdrstop

//---------------------------------------------------------------------------

#pragma argsused
void printBytes(CK_BYTE_PTR pBytes, DWORD dLenBytes) {

	for (DWORD d = 0; d < dLenBytes; d++) {
		printf("0x%X ",(pBytes[d]));
	}
}

void generateKeyPositive(CK_FUNCTION_LIST_PTR p11, CK_SESSION_HANDLE pHandle) {

   // Sign
   CK_MECHANISM pm;
   pm.mechanism = CKM_RSA_PKCS_KEY_PAIR_GEN;
   pm.pParameter = NULL;
   pm.ulParameterLen = 0;
   CK_BBOOL* bTRUE  = new CK_BBOOL(); *bTRUE  = CK_TRUE;
   CK_BBOOL* bFALSE = new CK_BBOOL(); *bFALSE = CK_FALSE;

   int iMaxPriv = 8;
   CK_ATTRIBUTE keyPrivTempl[]    = {
	{	CKA_SENSITIVE,			bTRUE,			sizeof(CK_BBOOL)	},
	{	CKA_DECRYPT,			bTRUE,			sizeof(CK_BBOOL)	},
	{	CKA_SIGN,				bTRUE,			sizeof(CK_BBOOL)	},
	{	CKA_SIGN_RECOVER,		bTRUE,			sizeof(CK_BBOOL)	},
	{	CKA_UNWRAP,				bFALSE,			sizeof(CK_BBOOL)	},
	{	CKA_EXTRACTABLE,		bFALSE,			sizeof(CK_BBOOL)	},
	{	CKA_ALWAYS_SENSITIVE,	bTRUE,			sizeof(CK_BBOOL)	},
	{	CKA_NEVER_EXTRACTABLE,	bTRUE,			sizeof(CK_BBOOL)	}
   };
   int iMaxPub = 5;
   CK_ATTRIBUTE keyPubTempl[]    = {
	{	CKA_ENCRYPT,			bTRUE,			sizeof(CK_BBOOL)	},
	{	CKA_VERIFY,				bTRUE,			sizeof(CK_BBOOL)	},
	{	CKA_VERIFY_RECOVER,		bTRUE,			sizeof(CK_BBOOL)	},
	{	CKA_WRAP,				bFALSE,			sizeof(CK_BBOOL)	},
	{	CKA_TRUSTED,			bFALSE,			sizeof(CK_BBOOL)	},
   };

   for (int i = 0; i < iMaxPriv; i++) {
	for (int k = 1; k < iMaxPub; k++) {
	   DWORD dPublic  = 0;
	   DWORD dPrivate = 0;
	   printf("\n %d Private key attributes, %d Public key attributes", i, k);
	   if (p11->C_GenerateKeyPair(pHandle,&pm,keyPubTempl,k,keyPrivTempl,i,&dPublic,&dPrivate) != CKR_OK) {
			printf("\nC_GenerateKeyPair failed !");
	   } else {
			if ( (dPublic != 0) && (dPrivate != 0)) {
				printf("\nC_GenerateKeyPair ok !");
			} else {
				printf("\nC_GenerateKeyPair failed !");
			}
	   }
	}
   }
}

int main(int argc, char* argv[])
{

   unsigned char ucPWDUser[] = { 0x31, 0x32, 0x33, 0x34, 0x35, 0x36 };
   DWORD dwLenPWDUser = 6;

   CK_FUNCTION_LIST_PTR p11;
   CK_RV returnCode;
   CK_C_GetFunctionList C_GetFunctionListDyn;

   if (argc != 2) {
		printf("Bitte LibName als Argument angeben");
		exit(0);
   }
   //DLL laden
   HINSTANCE hDll = LoadLibrary(argv[1]);
   if (hDll == NULL) {
		printf("Fehler beim Laden der Bibliothek");
		exit(0);
   }

   //Funktionen holen
   C_GetFunctionListDyn = (CK_C_GetFunctionList) GetProcAddress(hDll, "C_GetFunctionList");

   //Liste der Funktionen holen
   returnCode = C_GetFunctionListDyn(&p11);
   if (returnCode != CKR_OK) {
		FreeLibrary(hDll);
		hDll = NULL;
		p11 = NULL;
		printf("\nFehler bei C_GetFunctionList (dyn)");
		exit(0);
   }

   //PKCS#11 initialisieren
   returnCode = p11->C_Initialize(NULL_PTR);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Initialize");
		exit(0);
   } else {
        printf("\nC_Initialize erfolgreich");
   }

   //Anzahl Slots holen
   CK_ULONG ulSlotCount;
   returnCode = p11->C_GetSlotList(CK_TRUE, NULL_PTR, &ulSlotCount);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetSlotList");
        exit(0);
   } else {
        printf("\nC_GetSlotList erfolgreich");
   }

   //Sind Slots da?
   if (ulSlotCount == 0) {
        printf("\nKeine Slots gefunden");
        exit(0);
   } else {
		printf("\n %d Slots gefunden !",ulSlotCount);
   }

   //Slots holen
   CK_SLOT_ID_PTR pSlotId = (CK_SLOT_ID_PTR) malloc(sizeof(CK_SLOT_ID) * ulSlotCount);
   returnCode = p11->C_GetSlotList(CK_TRUE, pSlotId, &ulSlotCount);
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_GetSlotList");
        exit(0);
   } else {
        printf("\nC_GetSlotList erfolgreich, verwende SlotID : %X",pSlotId[0]);
   }

   //1. Slot verwenden !!
   //Session ffnen
   CK_SESSION_HANDLE hSession;
   returnCode = p11->C_OpenSession(pSlotId[0], (CKF_RW_SESSION | CKF_SERIAL_SESSION), "Test-InitPin", NULL_PTR, &hSession);
   //War das OK?
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_OpenSession");
        exit(0);
   } else {
		printf("\nC_OpenSession erfolgreich, Session Handle : %X",hSession);
   }

   //Einloggen ohne PinPad
   returnCode = p11->C_Login(hSession, CKU_USER, ucPWDUser, dwLenPWDUser);
   //War das OK?
   if (returnCode != CKR_OK) {
        printf("\nFehler bei C_Login");
        exit(0);
   } else {
		printf("\nC_Login erfolgreich, User: CKU_USER, PIN: '123456', Session: %X",hSession);
   }

   generateKeyPositive(p11,hSession);

   //Warten
   printf("\nTest erfolgreich !");
   getch();

}
//---------------------------------------------------------------------------
