//---------------------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <conio.h>

//Includes fr PKCS#11
#include "cryptoki.h"

#pragma hdrstop

//---------------------------------------------------------------------------

#pragma argsused
void printBytes(CK_BYTE_PTR pBytes, DWORD dLenBytes) {

	for (DWORD d = 0; d < dLenBytes; d++) {
		printf("0x%X ",(pBytes[d]));
	}
}

int main(int argc, char* argv[])
{


   bool bWithLogin = true;

   unsigned char ucPWDUser[] = { 0x31, 0x32, 0x33, 0x34, 0x35, 0x36 };
   DWORD dwLenPWDUser = 6;

   CK_FUNCTION_LIST_PTR p11;
   CK_RV returnCode = -1;
   CK_C_GetFunctionList C_GetFunctionListDyn;

   if (argc != 2) {
		printf("Bitte LibName als Argument angeben");
		exit(0);
   }
   //DLL laden
   HINSTANCE hDll = LoadLibrary(argv[1]);
   if (hDll == NULL) {
		printf("Error : %d",GetLastError());
		LPVOID lpMsgBuf;
		DWORD dw = GetLastError();

		FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			dw,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPTSTR) &lpMsgBuf,
			0, NULL );

		printf((const char *)lpMsgBuf);

		printf("Fehler beim Laden der Bibliothek");
		getch();
		exit(0);
   }

   //Funktionen holen
   C_GetFunctionListDyn = (CK_C_GetFunctionList) GetProcAddress(hDll, "C_GetFunctionList");

   //Liste der Funktionen holen
   if (C_GetFunctionListDyn != NULL) {
	   returnCode = C_GetFunctionListDyn(&p11);
   }

   if (returnCode != CKR_OK) {
		FreeLibrary(hDll);
		hDll = NULL;
		p11 = NULL;
		printf("\nFehler bei C_GetFunctionList (dyn)");
		getch();
		exit(0);
   }

   //PKCS#11 initialisieren
   returnCode = p11->C_Initialize(NULL_PTR);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Initialize");
		getch();
		exit(0);
   } else {
		printf("\nC_Initialize erfolgreich");
   }

   //Anzahl Slots holen
   CK_ULONG ulSlotCount;
   returnCode = p11->C_GetSlotList(CK_TRUE, NULL_PTR, &ulSlotCount);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetSlotList");
		getch();
		exit(0);
   } else {
		printf("\nC_GetSlotList erfolgreich");
   }

   //Sind Slots da?
   if (ulSlotCount == 0) {
		printf("\nKeine Slots gefunden");
		getch();
		exit(0);
   } else {
		printf("\n %d Slots gefunden !",ulSlotCount);
   }

   //Slots holen
   CK_SLOT_ID_PTR pSlotId = (CK_SLOT_ID_PTR) malloc(sizeof(CK_SLOT_ID) * ulSlotCount);
   returnCode = p11->C_GetSlotList(CK_TRUE, pSlotId, &ulSlotCount);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetSlotList");
		getch();
		exit(0);
   } else {
		printf("\nC_GetSlotList erfolgreich, verwende SlotID : %X",pSlotId[0]);
   }

   //Daten der Slots anzeigen
   for(unsigned long i=0; i<ulSlotCount; i++) {

	   CK_SLOT_INFO slotInfo;
	   returnCode = p11->C_GetSlotInfo(pSlotId[i], &slotInfo);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetSlotInfo");
			getch();
			exit(0);
	   }

	   //Infos zum Token holen
	   CK_TOKEN_INFO tokenInfo;
	   returnCode = p11->C_GetTokenInfo(pSlotId[i], &tokenInfo);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetTokenInfo");
			getch();
			exit(0);
	   }

	   //Daten ausgeben
	   printf("\n\nSlotId                     : %x", pSlotId[i]);
	   printf("\nBeschreibung des Slots       : ");
	   for (unsigned int i=0; i<64; i++) printf("%c", slotInfo.slotDescription[i]);
	   printf("\nHersteller des Slots         : ");
	   for(unsigned int i=0; i<32; i++) printf("%c", slotInfo.manufacturerID[i]);
	   printf("\nVersion der Hardware         : %d.%d", slotInfo.hardwareVersion.major, slotInfo.hardwareVersion.minor);
	   printf("\nVersion der Firmware         : %d.%d", slotInfo.firmwareVersion.major, slotInfo.firmwareVersion.minor);
	   printf("\nBezeichnung des Tokens       : ");
	   for (unsigned int i=0; i<32; i++) printf("%c", tokenInfo.label[i]);
	   printf("\nHersteller des Tokens        : ");
	   for (unsigned int i=0; i<32; i++) printf("%c", tokenInfo.manufacturerID[i]);
	   printf("\nModell des Tokens            : ");
	   for (unsigned int i=0; i<16; i++) printf("%c", tokenInfo.model[i]);
	   printf("\nSeriennummer des Tokens      : ");
	   for (unsigned int i=0; i<16; i++) printf("%c", tokenInfo.serialNumber[i]);
	   printf("\nMax. Sessions                : %d", tokenInfo.ulMaxSessionCount);
	   printf("\nVorhandene Sessions          : %d", tokenInfo.ulSessionCount);
	   printf("\nMax. Sessions (RW)           : %d", tokenInfo.ulMaxRwSessionCount);
	   printf("\nVorhandene Sessions (RW)     : %d", tokenInfo.ulRwSessionCount);
	   printf("\nMax. Pin-Laenge              : %d", tokenInfo.ulMaxPinLen);
	   printf("\nMin. Pin-Laenge              : %d", tokenInfo.ulMinPinLen);
	   printf("\nOeffentlicher Speicher       : %d", tokenInfo.ulTotalPublicMemory);
	   printf("\nFreier oeffentlicher Speicher: %d", tokenInfo.ulFreePublicMemory);
	   printf("\nPrivater Speicher            : %d", tokenInfo.ulTotalPrivateMemory);
	   printf("\nFreier privater Speicher     : %d", tokenInfo.ulFreePrivateMemory);
	   printf("\nVersion der Token-Hardware   : %d.%d", tokenInfo.hardwareVersion.major, slotInfo.hardwareVersion.minor);
	   printf("\nVersion der Token-Firmware   : %d.%d", tokenInfo.firmwareVersion.major, slotInfo.firmwareVersion.minor);

	   //Session ffnen
	   CK_SESSION_HANDLE hSession;
	   returnCode = p11->C_OpenSession(pSlotId[i], CKF_SERIAL_SESSION, "Test PKCS#11", NULL_PTR, &hSession);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_OpenSession");
			getch();
			exit(0);
	   }

	   if (bWithLogin) {
			//Einloggen ohne PinPad
			returnCode = p11->C_Login(hSession, CKU_USER, ucPWDUser, dwLenPWDUser);
			//War das OK?
			if (returnCode != CKR_OK) {
				printf("\nFehler bei C_Login");
				exit(0);
			} else {
				printf("\nC_Login erfolgreich, User: CKU_USER, PIN: '123456', Session: %X",hSession);
			}
	   }

	   //SessionInformation anzeigen
	   CK_SESSION_INFO sessionInfo;
	   returnCode = p11->C_GetSessionInfo(hSession, &sessionInfo);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_GetSessionInfo");
			getch();
			exit(0);
	   }

	   printf("\n\nSlotId (SessionInfo)    : %d", sessionInfo.slotID);
	   printf("\nFlags (SessionInfo)       : %d", sessionInfo.flags);
	   printf("\nStatus (SessionInfo)      : %d", sessionInfo.state);
	   printf("\nGertefehler (SessionInfo): %x", sessionInfo.ulDeviceError);

	   //Objekte des Tokens ausgeben
	   returnCode = p11->C_FindObjectsInit(hSession, NULL_PTR, 0);
	   if (returnCode != CKR_OK) {
			printf("\nFehler bei C_FindObjectsInit");
			getch();
			exit(0);
	   }

	   //Suchen
	   CK_OBJECT_HANDLE objectList[20];
	   CK_ULONG objectsFound=0;

	   returnCode = p11->C_FindObjects(hSession, objectList, 20, &objectsFound);
	   if (returnCode != CKR_OK) {
            printf("\nFehler bei C_FindObjects");
			getch();
			exit(0);
	   }

	   returnCode = p11->C_FindObjectsFinal(hSession);
	   if (returnCode != CKR_OK) {
            printf("\nFehler bei C_FindObjectsFinal");
			getch();
			exit(0);
	   }

	   //Jetzt die Daten aller Objekte anzeigen
	   printf("\n\nObjekte des Tokens: ");

	   for(int i=0; i<objectsFound; i++) {

             CK_ATTRIBUTE attr[20];
             char objectLabel[100];

             //Nur den Label holen
             attr[0].type = CKA_LABEL;
			 attr[0].pValue = objectLabel;
			 attr[0].ulValueLen = sizeof(objectLabel);

             //Attribute holen
			 returnCode = p11->C_GetAttributeValue(hSession, objectList[i], attr, 1);
			 if (returnCode != CKR_OK) {
				 printf("\nFehler bei C_GetAttributeValue");
				getch();
				exit(0);
			 }
             //Ausgeben
             printf("\n");
             for(int j=0; j<attr[0].ulValueLen; j++) printf("%c", objectLabel[j]);
	   }

	   //Session beenden
	   returnCode = p11->C_CloseSession(hSession);

	   if (returnCode != CKR_OK) {
            printf("\nFehler bei C_CloseSession");
			getch();
			exit(0);
	   }
   }

   //Warten
   printf("\nTest erfolgreich !");
   getch();

   //Sitzung beenden
   returnCode = p11->C_Finalize(NULL_PTR);

   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Finalize");
		getch();
		exit(0);
   }
}
//---------------------------------------------------------------------------
