//---------------------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <conio.h>

//Includes fr PKCS#11
#include "cryptoki.h"

#pragma hdrstop

//---------------------------------------------------------------------------

#pragma argsused
void printBytes(CK_BYTE_PTR pBytes, DWORD dLenBytes) {

	for (DWORD d = 0; d < dLenBytes; d++) {
		printf("0x%X ",(pBytes[d]));
	}
}

int main(int argc, char* argv[])
{

   unsigned char ucPWDUser[] = { 0x31, 0x32, 0x33, 0x34, 0x35, 0x36 };
   DWORD dwLenPWDUser = 6;

   CK_FUNCTION_LIST_PTR p11;
   CK_RV returnCode;
   CK_C_GetFunctionList C_GetFunctionListDyn;

   if (argc != 2) {
		printf("Bitte LibName als Argument angeben");
		exit(0);
   }
   //DLL laden
   HINSTANCE hDll = LoadLibrary(argv[1]);
   if (hDll == NULL) {
		printf("Fehler beim Laden der Bibliothek");
		exit(0);
   }

   //Funktionen holen
   C_GetFunctionListDyn = (CK_C_GetFunctionList) GetProcAddress(hDll, "C_GetFunctionList");

   //Liste der Funktionen holen
   returnCode = C_GetFunctionListDyn(&p11);
   if (returnCode != CKR_OK) {
		FreeLibrary(hDll);
		hDll = NULL;
		p11 = NULL;
		printf("\nFehler bei C_GetFunctionList (dyn)");
		exit(0);
   }

   //PKCS#11 initialisieren
   returnCode = p11->C_Initialize(NULL_PTR);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Initialize");
		getch();
		exit(0);
   } else {
		printf("\nC_Initialize erfolgreich");
   }

   //Anzahl Slots holen
   CK_ULONG ulSlotCount;
   returnCode = p11->C_GetSlotList(CK_TRUE, NULL_PTR, &ulSlotCount);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetSlotList");
		getch();
		exit(0);
   } else {
		printf("\nC_GetSlotList erfolgreich");
   }

   //Sind Slots da?
   if (ulSlotCount == 0) {
		printf("\nKeine Slots gefunden");
		getch();
		exit(0);
   } else {
		printf("\n %d Slots gefunden !",ulSlotCount);
   }

   //Slots holen
   CK_SLOT_ID_PTR pSlotId = (CK_SLOT_ID_PTR) malloc(sizeof(CK_SLOT_ID) * ulSlotCount);
   returnCode = p11->C_GetSlotList(CK_TRUE, pSlotId, &ulSlotCount);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetSlotList");
		getch();
		exit(0);
   } else {
		printf("\nC_GetSlotList erfolgreich, verwende SlotID : %X",pSlotId[0]);
   }

   //1. Slot verwenden !!
   //Session ffnen
   CK_SESSION_HANDLE hSession;
   returnCode = p11->C_OpenSession(pSlotId[0], CKF_SERIAL_SESSION, "Test-InitPin", NULL_PTR, &hSession);
   //War das OK?
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_OpenSession");
		getch();
		exit(0);
   } else {
		printf("\nC_OpenSession erfolgreich, Session Handle : %X",hSession);
   }

   // Auf Protected_Authentication_Path berprfen
   //Infos zum Token holen
   CK_TOKEN_INFO tokenInfo;
   returnCode = p11->C_GetTokenInfo(pSlotId[0], &tokenInfo);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetTokenInfo");
		getch();
		exit(0);
   }

   if ( (tokenInfo.flags & CKF_PROTECTED_AUTHENTICATION_PATH) == 0x00) {
		printf("\nSlot verfgt nicht ber PinPad !");
		getch();
		exit(0);
   }

   //Einloggen mit PinPad
   printf("\nPin am PinPad eingeben:");
   returnCode = p11->C_Login(hSession, CKU_USER, NULL, 0);

   //War das OK?
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_Login");
		getch();
		exit(0);
   } else {
		printf("\nC_Login erfolgreich, User: CKU_USER, PIN: ???, Session: %X",hSession);
   }

   //////////////////////////////////////
   // Zertifikat suchen
   // Template fr X.509 Zertifikat
   CK_OBJECT_CLASS cClass   = CKO_CERTIFICATE;
   CK_CERTIFICATE_TYPE   certType  = CKC_X_509;
   CK_ATTRIBUTE certTempl[]    = {
		{
			CKA_CLASS,
			(unsigned char *) &cClass,
			sizeof(cClass)
		},
		{
			CKA_CERTIFICATE_TYPE,
			(unsigned char *) &certType,
			sizeof(CK_CERTIFICATE_TYPE)
		}
   };

   returnCode = p11->C_FindObjectsInit(hSession,certTempl,2);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_FindObjectsInit");
		getch();
		exit(0);
   } else {
		printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_CERTIFICATE, CKA_CERTIFICATE_TYPE: CKC_X_509");
   }
   CK_OBJECT_HANDLE pCertificateHandle;
   CK_ULONG ulMaxCount;

   // Erstes gefundene Zertifikat verwenden in pCertificateHandle
   returnCode = p11->C_FindObjects(hSession,&pCertificateHandle,1,&ulMaxCount);
   if ( (returnCode != CKR_OK) || (ulMaxCount==0)) {
		printf("\nFehler bei C_FindObjects");
		getch();
		exit(0);
   } else {
		printf("\nC_FindObjects erfolgreich, CERTIFICATE Handle: %X", pCertificateHandle);
   }

   // Suche abschliessen
   returnCode = p11->C_FindObjectsFinal(hSession);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_FindObjectsFinal");
		getch();
		exit(0);
   } else {
		printf("\nC_FindObjectsFinal erfolgreich");
   }

   // Zertifikatshandle nun in pCertificateHandle !!
   // Anhand dessen CKA_ID finden
   CK_BYTE pbCKA_ID[1024];
   memset(pbCKA_ID,0x00,1024);
   CK_ATTRIBUTE paKeyCKA_ID[]   = { { CKA_ID, NULL, 0 } };
   p11->C_GetAttributeValue(hSession, pCertificateHandle, paKeyCKA_ID, 1);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetAttributeValue");
		getch();
		exit(0);
   } else {
		printf("\nC_GetAttributeValue erfolgreich, CKA_ID Lnge: %d", paKeyCKA_ID[0].ulValueLen);
   }
   DWORD dLenID = paKeyCKA_ID[0].ulValueLen;
   CK_ATTRIBUTE pKeyCKA_ID[] = { { CKA_ID, pbCKA_ID, dLenID } };
   p11->C_GetAttributeValue(hSession, pCertificateHandle, pKeyCKA_ID, 1);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_GetAttributeValue");
		getch();
		exit(0);
   } else {
		printf("\nC_GetAttributeValue erfolgreich, CKA_ID :");
		printBytes(pbCKA_ID,dLenID);
		printf("\n");
   }

   //////////////////////////////////////
   // PrivateKey anhand der CKA_ID des Zertifikats finden
   cClass = CKO_PRIVATE_KEY;
   CK_ATTRIBUTE keyTempl[]    = {
		{
			CKA_CLASS,
			(unsigned char *) &cClass,
			sizeof(cClass)
		},
		{
			CKA_ID,
		   (unsigned char *) &pbCKA_ID,
			dLenID
		}
   };

   returnCode = p11->C_FindObjectsInit(hSession,keyTempl,2);
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_FindObjectsInit");
		getch();
		exit(0);
   } else {
		printf("\nC_FindObjectsInit erfolgreich, CKA_CLASS: CKO_PRIVATE_KEY CKA_ID: ");
		printBytes(pbCKA_ID,dLenID);
		printf("\n");
   }

   CK_OBJECT_HANDLE pObject;
   ulMaxCount;
   returnCode = p11->C_FindObjects(hSession,&pObject,1,&ulMaxCount);
   if ( (returnCode != CKR_OK) || (ulMaxCount==0))  {
		printf("\nFehler bei C_FindObjects");
		if (ulMaxCount==0)
			printf("\nKeinen privaten Schlssel zum Zertifikat gefunden !");
		getch();
		exit(0);
   } else {
		printf("\nC_FindObjects erfolgreich, Key Handle: %X", pObject);
   }

   returnCode = p11->C_FindObjectsFinal(hSession);

   //War das OK?
   if (returnCode != CKR_OK) {
		printf("\nFehler bei C_FindObjectsFinal");
		getch();
		exit(0);
   } else {
		printf("\nC_FindObjectsFinal erfolgreich");
   }

   //////////////////////////////////////
   //Signatur beginnen
   CK_MECHANISM mechanism;
   CK_OBJECT_HANDLE hKey = pObject;
   CK_BYTE data[1024], data2[1024], sig[256];
   CK_ULONG lenSig=256;
   mechanism.mechanism = CKM_SHA512_RSA_PKCS;
   mechanism.pParameter = NULL_PTR;
   mechanism.ulParameterLen = 0;

   for(int i=0; i<1024; i++) data[i] = i % 256;

   // Signaturvorgang initialisieren
   returnCode = p11->C_SignInit(hSession, &mechanism, hKey);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei C_SignInit");
		p11->C_CloseSession(hSession);
		free(pSlotId);
		p11->C_Finalize(NULL_PTR);
		getch();
		exit(0);
	 }
   printf("\nC_SignInit erfolgreich");

   //Signieren
   returnCode = p11->C_Sign(hSession, data, 1024, sig, &lenSig);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei C_Sign");
		p11->C_CloseSession(hSession);
		free(pSlotId);
		p11->C_Finalize(NULL_PTR);
		getch();
		exit(0);
	 }
   printf("\nC_Sign erfolgreich !\n Signatur:\n");
   printBytes(sig,lenSig);
   printf("\n");

   //Session schlieen
   returnCode = p11->C_CloseSession(hSession);

   //War das OK?
   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei C_CloseSession");
		free(pSlotId);
		p11->C_Finalize(NULL_PTR);
		getch();
		exit(0);
	 }

   //Speicher freigeben
   free(pSlotId);

   //Sitzung beenden
   returnCode = p11->C_Finalize(NULL_PTR);

   if (returnCode != CKR_OK)
	 {
		printf("\nFehler bei C_Finalize");
		getch();
		exit(0);
	 }

   //Warten
   printf("\nTest erfolgreich !");
   getch();

}
//---------------------------------------------------------------------------
